from flask import current_app
from flask_jwt_extended import jwt_required

from zou.app.models.task_type import TaskType
from zou.app.models.task_status import TaskStatus
from zou.app.models.project import Project
from zou.app.models.person import Person
from zou.app.models.task import Task

from zou.app.services import deletion_service, tasks_service

from zou.app.blueprints.source.shotgun.base import (
    BaseImportShotgunResource,
    ImportRemoveShotgunBaseResource,
)


class ImportShotgunTasksResource(BaseImportShotgunResource):
    @jwt_required()
    def post(self):
        """
        Import shotgun tasks
        ---
        description: Import Shotgun tasks. Send a list of Shotgun task
          entries in the JSON body. Only tasks with steps and projects are
          imported. Returns created or updated tasks with assignees.
        tags:
          - Import
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: array
                items:
                  type: object
                  properties:
                    id:
                      type: integer
                      description: Shotgun ID of the task
                      example: 12345
                    cached_display_name:
                      type: string
                      description: Task display name
                      example: "Modeling for Asset"
                    start_date:
                      type: string
                      format: date
                      description: Start date
                      example: "2024-01-15"
                    due_date:
                      type: string
                      format: date
                      description: Due date
                      example: "2024-02-15"
                    sg_sort_order:
                      type: integer
                      description: Sort order
                      example: 1
                    duration:
                      type: integer
                      description: Duration in days
                      example: 30
                    step:
                      type: object
                      description: Step (task type) information
                      properties:
                        id:
                          type: integer
                          example: 11111
                    project:
                      type: object
                      description: Project information
                      properties:
                        id:
                          type: integer
                          example: 22222
                    entity:
                      type: object
                      description: Entity information
                      properties:
                        id:
                          type: integer
                          example: 33333
                        type:
                          type: string
                          example: "Asset"
                    sg_status_list:
                      type: string
                      description: Status short name
                      example: "IP"
                    created_by:
                      type: object
                      description: Creator information
                      properties:
                        id:
                          type: integer
                          example: 44444
                    task_assignees:
                      type: array
                      description: Task assignees
                      items:
                        type: object
                        properties:
                          id:
                            type: integer
                            example: 55555
              example:
                - id: 12345
                  cached_display_name: "Modeling for Asset"
                  start_date: "2024-01-15"
                  due_date: "2024-02-15"
                  sg_sort_order: 1
                  duration: 30
                  step:
                    id: 11111
                  project:
                    id: 22222
                  entity:
                    id: 33333
                    type: "Asset"
                  sg_status_list: "IP"
                  created_by:
                    id: 44444
                  task_assignees:
                    - id: 55555
        responses:
          200:
            description: Tasks imported successfully
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Task unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        description: Task name
                        example: "Modeling for Asset"
                      start_date:
                        type: string
                        format: date
                        description: Start date
                        example: "2024-01-15"
                      due_date:
                        type: string
                        format: date
                        description: Due date
                        example: "2024-02-15"
                      created_at:
                        type: string
                        format: date-time
                        description: Creation timestamp
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        description: Update timestamp
                        example: "2024-01-15T11:00:00Z"
          400:
            description: Invalid request body or data format error
        """
        return super().post()

    def prepare_import(self):
        self.project_ids = Project.get_id_map()
        self.person_ids = Person.get_id_map()
        self.task_type_ids = TaskType.get_id_map(field="shotgun_id")
        self.task_status_ids = TaskStatus.get_id_map(field="short_name")

    def filtered_entries(self):
        return [x for x in self.sg_entries if self.is_valid_task(x)]

    def is_valid_task(self, sg_task):
        return sg_task["step"] is not None and sg_task["project"] is not None

    def extract_data(self, sg_task):
        entity_id = self.get_entity_id(sg_task["entity"])
        task_status_id = self.task_status_ids.get(
            sg_task["sg_status_list"], None
        )
        step_shotgun_id = sg_task["step"]["id"]
        assigner_id = self.person_ids.get(sg_task["created_by"]["id"], None)
        project_id = self.project_ids.get(sg_task["project"]["id"], None)
        task_type_id = self.task_type_ids.get(step_shotgun_id, None)
        assignees = self.extract_assignees(sg_task, self.person_ids)

        return {
            "name": sg_task["cached_display_name"],
            "shotgun_id": sg_task["id"],
            "start_date": sg_task["start_date"],
            "due_date": sg_task["due_date"],
            "sort_order": sg_task["sg_sort_order"],
            "duration": sg_task["duration"],
            "task_type_id": task_type_id,
            "task_status_id": task_status_id,
            "project_id": project_id,
            "entity_id": entity_id,
            "assigner_id": assigner_id,
            "assignees": assignees,
        }

    def get_entity_id(self, sg_entity):
        entity_id = None
        if sg_entity is not None:
            entity_sg_id = sg_entity["id"]
            entity_type = sg_entity["type"]
            if entity_type == "Asset":
                entity_id = self.get_asset_id(entity_sg_id)
            elif entity_type == "Shot":
                entity_id = self.get_shot_id(entity_sg_id)
            elif entity_type == "Scene":
                entity_id = self.get_scene_id(entity_sg_id)
            elif entity_type == "Sequence":
                entity_id = self.get_sequence_id(entity_sg_id)
            elif entity_type == "Episode":
                entity_id = self.get_episode_id(entity_sg_id)
            else:
                return None

        return entity_id

    def extract_assignees(self, sg_task, person_ids):
        assignees = []
        if len(sg_task["task_assignees"]) > 0:
            for sg_person in sg_task["task_assignees"]:
                person_id = person_ids[sg_person["id"]]
                person = Person.get(person_id)
                assignees.append(person)
        return assignees

    def import_entry(self, data):
        task = Task.get_by(shotgun_id=data["shotgun_id"])

        if task is None:
            task = Task.get_by(
                name=data["name"],
                project_id=data["project_id"],
                task_type_id=data["task_type_id"],
                entity_id=data["entity_id"],
            )

        if task is None:
            task = Task(**data)
            task.save()
            current_app.logger.info("Task created: %s" % task)
        else:
            existing_task = Task.get_by(
                name=data["name"],
                project_id=data["project_id"],
                task_type_id=data["task_type_id"],
                entity_id=data["entity_id"],
            )

            if existing_task is not None:
                data.pop("name", None)
                data.pop("project_id", None)
                data.pop("task_type_id", None)
                data.pop("entity_id", None)

            task.update(data)
            tasks_service.clear_task_cache(str(task.id))
            current_app.logger.info("Task updated: %s" % task)

        return task


class ImportRemoveShotgunTaskResource(ImportRemoveShotgunBaseResource):
    def __init__(self):
        ImportRemoveShotgunBaseResource.__init__(self, Task, self.delete_func)

    @jwt_required()
    def post(self):
        """
        Remove shotgun task
        ---
        description: Remove a Shotgun task from the database. Provide the
          Shotgun entry ID in the JSON body.
        tags:
          - Import
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - id
                properties:
                  id:
                    type: integer
                    description: Shotgun ID of the task to remove
                    example: 12345
              example:
                id: 12345
        responses:
          200:
            description: Removal result returned
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    success:
                      type: boolean
                      description: Whether the removal was successful
                      example: true
                    removed_instance_id:
                      type: string
                      format: uuid
                      description: ID of the removed task, if found
                      example: a24a6ea4-ce75-4665-a070-57453082c25
          400:
            description: Invalid request body or instance not found
        """
        return super().post()

    def delete_func(self, entity):
        deletion_service.remove_task(entity.id)
        tasks_service.clear_task_cache(str(entity.id))



from zou.app import config
from zou.app.utils import emails, chats
from zou.app.utils.email_i18n import get_email_translation

from zou.app.services import (
    entities_service,
    names_service,
    persons_service,
    projects_service,
    shots_service,
    tasks_service,
)
from zou.app.stores import queue_store
from zou.app.services.templates_service import generate_html_body


def send_notification(
    person_id, subject, messages, title="", force_email=False, locale=None
):
    """
    Send email notification to given person. Use the job queue if it is
    activated. If locale is provided, the email Content-Language header
    is set accordingly.
    """
    person = persons_service.get_person(person_id)
    email_message = messages["email_message"]
    slack_message = messages["slack_message"]
    mattermost_message = messages["mattermost_message"]
    discord_message = messages["discord_message"]
    email_locale = (
        locale or person.get("locale") or persons_service.get_default_locale()
    )
    email_html_body = generate_html_body(
        title, email_message, locale=email_locale
    )

    if person["notifications_enabled"] or force_email:
        if config.ENABLE_JOB_QUEUE:
            queue_store.job_queue.enqueue(
                emails.send_email,
                args=(
                    subject,
                    email_html_body,
                    person["email"],
                ),
                kwargs={"locale": email_locale},
            )
        else:
            emails.send_email(
                subject, email_html_body, person["email"], locale=email_locale
            )

    if person["notifications_slack_enabled"]:
        organisation = persons_service.get_organisation(sensitive=True)
        userid = person["notifications_slack_userid"]
        token = organisation.get("chat_token_slack", "")
        if config.ENABLE_JOB_QUEUE:
            queue_store.job_queue.enqueue(
                chats.send_to_slack,
                args=(token, userid, slack_message),
            )
        else:
            chats.send_to_slack(token, userid, slack_message)

    if person["notifications_mattermost_enabled"]:
        organisation = persons_service.get_organisation(sensitive=True)
        userid = person["notifications_mattermost_userid"]
        webhook = organisation.get("chat_webhook_mattermost", "")
        if config.ENABLE_JOB_QUEUE:
            queue_store.job_queue.enqueue(
                chats.send_to_mattermost,
                args=(webhook, userid, mattermost_message),
            )
        else:
            chats.send_to_mattermost(webhook, userid, mattermost_message)

    if person["notifications_discord_enabled"]:
        organisation = persons_service.get_organisation(sensitive=True)
        userid = person["notifications_discord_userid"]
        token = organisation.get("chat_token_discord", "")
        if config.ENABLE_JOB_QUEUE:
            queue_store.job_queue.enqueue(
                chats.send_to_discord,
                args=(token, userid, discord_message),
            )
        else:
            chats.send_to_discord(token, userid, discord_message)

    return True


def send_comment_notification(person_id, author_id, comment, task):
    """
    Send a notification email telling that a new comment was posted to person
    matching given person id. Email content is translated according to the
    recipient's locale.
    """
    person = persons_service.get_person(person_id)
    project = projects_service.get_project(task["project_id"])
    locale = person.get("locale") or persons_service.get_default_locale()
    if (
        person["notifications_enabled"]
        or person["notifications_slack_enabled"]
        or person["notifications_mattermost_enabled"]
        or person["notifications_discord_enabled"]
    ):
        task_status = tasks_service.get_task_status(task["task_status_id"])
        task_status_name = task_status["short_name"].upper()
        author, task_name, task_url = get_task_descriptors(author_id, task)
        subject = get_email_translation(
            locale,
            "comment_subject",
            task_status_name=task_status_name,
            author_first_name=author["first_name"],
            task_name=task_name,
        )
        email_params = {
            "author_full_name": author["full_name"],
            "task_url": task_url,
            "task_name": task_name,
            "task_status_name": task_status_name,
        }
        if len(comment["text"]) > 0:
            email_message = get_email_translation(
                locale,
                "comment_body_with_text",
                comment_text=comment["text"],
                **email_params,
            )
            slack_message = """*%s* wrote a comment on <%s|%s> and set the status to *%s*.

_%s_
""" % (
                author["full_name"],
                task_url,
                task_name,
                task_status_name,
                comment["text"],
            )

            discord_message = """*%s* wrote a comment on [%s](%s)> and set the status to *%s*.

_%s_
""" % (
                author["full_name"],
                task_name,
                task_url,
                task_status_name,
                comment["text"],
            )

        else:
            email_message = get_email_translation(
                locale, "comment_body_status_only", **email_params
            )
            slack_message = """*%s* changed status of <%s|%s> to *%s*.
""" % (
                author["full_name"],
                task_url,
                task_name,
                task_status_name,
            )

            discord_message = """*%s* changed status of [%s](%s) to *%s*.
""" % (
                author["full_name"],
                task_name,
                task_url,
                task_status_name,
            )

        title = get_email_translation(locale, "comment_title")
        messages = {
            "email_message": email_message,
            "slack_message": slack_message,
            "mattermost_message": {
                "message": slack_message,
                "project_name": project["name"],
            },
            "discord_message": discord_message,
        }
        send_notification(person_id, subject, messages, title)

    return True


def send_mention_notification(person_id, author_id, comment, task):
    """
    Send a notification email telling that somenone mentioned the
    person matching given person id. Email content is translated
    according to the recipient's locale.
    """
    person = persons_service.get_person(person_id)
    project = projects_service.get_project(task["project_id"])
    locale = person.get("locale") or persons_service.get_default_locale()
    if (
        person["notifications_enabled"]
        or person["notifications_slack_enabled"]
        or person["notifications_mattermost_enabled"]
        or person["notifications_discord_enabled"]
    ):
        author, task_name, task_url = get_task_descriptors(author_id, task)
        subject = get_email_translation(
            locale,
            "mention_subject",
            author_first_name=author["first_name"],
            task_name=task_name,
        )
        email_message = get_email_translation(
            locale,
            "mention_body",
            author_full_name=author["full_name"],
            task_url=task_url,
            task_name=task_name,
            comment_text=comment["text"],
        )
        slack_message = """*%s* mentioned you in a comment on <%s|%s>.

_%s_
""" % (
            author["full_name"],
            task_url,
            task_name,
            comment["text"],
        )

        discord_message = """*%s* mentioned you in a comment on [%s](%s).

_%s_
""" % (
            author["full_name"],
            task_name,
            task_url,
            comment["text"],
        )
        title = get_email_translation(locale, "mention_title")
        messages = {
            "email_message": email_message,
            "slack_message": slack_message,
            "mattermost_message": {
                "message": slack_message,
                "project_name": project["name"],
            },
            "discord_message": discord_message,
        }
        return send_notification(person_id, subject, messages, title)
    else:
        return True


def send_assignation_notification(person_id, author_id, task):
    """
    Send a notification email telling that somenone assigned to a task the
    person matching given person id. Email content is translated according
    to the recipient's locale.
    """
    person = persons_service.get_person(person_id)
    project = projects_service.get_project(task["project_id"])
    locale = person.get("locale") or persons_service.get_default_locale()
    if (
        person["notifications_enabled"]
        or person["notifications_slack_enabled"]
        or person["notifications_mattermost_enabled"]
        or person["notifications_discord_enabled"]
    ):
        author, task_name, task_url = get_task_descriptors(author_id, task)
        subject = get_email_translation(
            locale, "assignation_subject", task_name=task_name
        )
        email_message = get_email_translation(
            locale,
            "assignation_body",
            author_full_name=author["full_name"],
            task_url=task_url,
            task_name=task_name,
        )
        slack_message = """*%s* assigned you to <%s|%s>.
""" % (
            author["full_name"],
            task_url,
            task_name,
        )
        discord_message = """*%s* assigned you to [%s](%s).
""" % (
            author["full_name"],
            task_name,
            task_url,
        )

        title = get_email_translation(locale, "assignation_title")
        messages = {
            "email_message": email_message,
            "slack_message": slack_message,
            "mattermost_message": {
                "message": slack_message,
                "project_name": project["name"],
            },
            "discord_message": discord_message,
        }
        return send_notification(person_id, subject, messages, title)
    return True


def get_task_descriptors(person_id, task):
    """
    Build task information needed to write notification emails: author object,
    full task name and task URL.
    """
    author = persons_service.get_person(person_id)
    project = projects_service.get_project(task["project_id"])
    task_type = tasks_service.get_task_type(task["task_type_id"])
    entity = entities_service.get_entity(task["entity_id"])
    entity_name, episode_id, _ = names_service.get_full_entity_name(
        entity["id"]
    )

    episode_segment = ""
    entity_type = "assets"
    if task_type["for_entity"] == "Shot":
        entity_type = "shots"
    if task_type["for_entity"] == "Edit":
        entity_type = "edits"
    if project["production_type"] == "tvshow":
        episode_segment = "/episodes/%s" % episode_id

    task_name = "%s / %s / %s" % (
        project["name"],
        entity_name,
        task_type["name"],
    )
    task_url = "%s://%s/productions/%s%s/%s/tasks/%s" % (
        config.DOMAIN_PROTOCOL,
        config.DOMAIN_NAME,
        task["project_id"],
        episode_segment,
        entity_type,
        task["id"],
    )
    return (author, task_name, task_url)


def send_reply_notification(person_id, author_id, comment, task, reply):
    """
    Send a notification email telling that a new reply was posted to person
    matching given person id. Email content is translated according to the
    recipient's locale.
    """
    person = persons_service.get_person(person_id)
    locale = person.get("locale") or persons_service.get_default_locale()
    if (
        person["notifications_enabled"]
        or person["notifications_slack_enabled"]
        or person["notifications_mattermost_enabled"]
        or person["notifications_discord_enabled"]
    ):
        tasks_service.get_task_status(task["task_status_id"])
        project = projects_service.get_project(task["project_id"])
        author, task_name, task_url = get_task_descriptors(author_id, task)
        subject = get_email_translation(
            locale,
            "reply_subject",
            author_first_name=author["first_name"],
            task_name=task_name,
        )
        email_message = get_email_translation(
            locale,
            "reply_body",
            author_full_name=author["full_name"],
            task_url=task_url,
            task_name=task_name,
            reply_text=reply["text"],
        )
        slack_message = """*%s* wrote a reply on <%s|%s>.

_%s_
""" % (
            author["full_name"],
            task_url,
            task_name,
            reply["text"],
        )

        discord_message = """*%s* wrote a reply on [%s](%s).

_%s_
""" % (
            author["full_name"],
            task_name,
            task_url,
            reply["text"],
        )

        title = get_email_translation(locale, "reply_title")
        messages = {
            "email_message": email_message,
            "slack_message": slack_message,
            "mattermost_message": {
                "message": slack_message,
                "project_name": project["name"],
            },
            "discord_message": discord_message,
        }
        send_notification(person_id, subject, messages, title)
    return True


def send_playlist_ready_notification(person_id, author_id, playlist):
    """
    Send a notification email telling that a new playlist is ready to person
    matching given person id. Email content is translated according to the
    recipient's locale.
    """
    person = persons_service.get_person(person_id)
    author = persons_service.get_person(author_id)
    project = projects_service.get_project(playlist["project_id"])
    locale = person.get("locale") or persons_service.get_default_locale()
    episode = None
    try:
        episode = shots_service.get_episode(playlist["episode_id"])
    except Exception:
        pass

    if (
        person["notifications_enabled"]
        or person["notifications_slack_enabled"]
        or person["notifications_mattermost_enabled"]
        or person["notifications_discord_enabled"]
    ):

        playlist_url = f"{config.DOMAIN_PROTOCOL}://{config.DOMAIN_NAME}/productions/{playlist['project_id']}/"

        if episode is not None:
            playlist_url += (
                f"episodes/{episode['id']}/playlists/{playlist['id']}"
            )
        elif (
            project["production_type"] == "tvshow"
            and playlist["for_entity"] == "asset"
        ):
            if playlist["is_for_all"] == True:
                playlist_url += f"episodes/all/playlists/{playlist['id']}"
            else:
                playlist_url += f"episodes/main/playlists/{playlist['id']}"
        else:
            playlist_url += f"playlists/{playlist['id']}"

        if episode is not None:
            episode_segment = get_email_translation(
                locale,
                "playlist_episode_segment",
                episode_name=episode["name"],
            )
        else:
            episode_segment = ""

        title = get_email_translation(locale, "playlist_title")
        subject = get_email_translation(
            locale,
            "playlist_subject",
            playlist_name=playlist["name"],
            project_name=project["name"],
        )

        email_message = get_email_translation(
            locale,
            "playlist_body",
            author_full_name=author["full_name"],
            playlist_url=playlist_url,
            playlist_name=playlist["name"],
            episode_segment=episode_segment,
            project_name=project["name"],
        )

        if len(playlist["shots"]) > 1:
            email_message += get_email_translation(
                locale,
                "playlist_elements_count",
                count=len(playlist["shots"]),
            )

        slack_message = f"*{author['full_name']}* notifies you that a playlist <{playlist_url}|{playlist['name']}> is ready for a review under {episode_segment}the project {project['name']}."

        discord_message = f"*{author['full_name']}* notifies you that a playlist [{playlist['name']}]({playlist_url}) is ready for a review under {episode_segment}the project {project['name']}."
        messages = {
            "email_message": email_message,
            "slack_message": slack_message,
            "mattermost_message": {
                "message": slack_message,
                "project_name": project["name"],
            },
            "discord_message": discord_message,
        }
        send_notification(
            person_id, subject, messages, title, force_email=True
        )



import slugify

import os

from flask import abort, after_this_request, request, send_file as flask_send_file
from flask_restful import Resource
from flask_jwt_extended import jwt_required

from zou.app import config
from zou.app.mixin import ArgsMixin

from zou.app.blueprints.playlists.schemas import (
    AddEntityToPlaylistSchema,
    NotifyClientsPlaylistSchema,
    TempPlaylistCreateSchema,
)
from zou.app.services import (
    entities_service,
    notifications_service,
    playlists_service,
    persons_service,
    preview_files_service,
    projects_service,
    shots_service,
    user_service,
)
from zou.app.stores import file_store, queue_store
from zou.app.utils import fs, permissions, validation
from zou.utils.movie import EncodingParameters


class ProjectPlaylistsResource(Resource, ArgsMixin):

    @jwt_required()
    def get(self, project_id):
        """
        Get project playlists
        ---
        description: Retrieve all playlists related to given project. Result is
          paginated and can be sorted.
        tags:
          - Playlists
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: page
            required: false
            schema:
              type: integer
            description: Page number for pagination
            example: 1
          - in: query
            name: sort_by
            required: false
            schema:
              type: string
            description: Field to sort by
            example: updated_at
          - in: query
            name: task_type_id
            required: false
            schema:
              type: string
              format: uuid
            description: Task type unique identifier to filter by
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          200:
            description: All playlists related to given project
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Playlist unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        description: Playlist name
                        example: "Review Playlist"
                      project_id:
                        type: string
                        format: uuid
                        description: Project unique identifier
                        example: b35b7fb5-df86-5776-b181-68564193d36
        """
        user_service.block_access_to_vendor()
        user_service.check_project_access(project_id)
        page = self.get_page()
        sort_by = self.get_sort_by()
        task_type_id = self.get_text_parameter("task_type_id")
        return playlists_service.all_playlists_for_project(
            project_id,
            for_client=permissions.has_client_permissions(),
            page=page,
            sort_by=sort_by,
            task_type_id=task_type_id,
        )


class EpisodePlaylistsResource(Resource, ArgsMixin):

    @jwt_required()
    def get(self, project_id, episode_id):
        """
        Get episode playlists
        ---
        description: Retrieve all playlists related to given episode. The full
          list is returned because the number of playlists in an episode is not
          that big.
        tags:
          - Playlists
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: episode_id
            required: true
            schema:
              type: string
              format: uuid
            description: Episode unique identifier or special value (main, all)
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: All playlists related to given episode
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Playlist unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        description: Playlist name
                        example: "Review Playlist"
                      episode_id:
                        type: string
                        format: uuid
                        description: Episode unique identifier
                        example: b35b7fb5-df86-5776-b181-68564193d36
        """
        user_service.block_access_to_vendor()
        user_service.check_project_access(project_id)
        page = self.get_page()
        sort_by = self.get_sort_by()
        task_type_id = self.get_text_parameter("task_type_id")
        if episode_id not in ["main", "all"]:
            shots_service.get_episode(episode_id)
        return playlists_service.all_playlists_for_episode(
            project_id,
            episode_id,
            for_client=permissions.has_client_permissions(),
            page=page,
            sort_by=sort_by,
            task_type_id=task_type_id,
        )


class ProjectPlaylistResource(Resource):

    @jwt_required()
    def get(self, project_id, playlist_id):
        """
        Get playlist
        ---
        description: Retrieve a specific playlist by ID with preview file
          revisions.
        tags:
          - Playlists
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: playlist_id
            required: true
            schema:
              type: string
              format: uuid
            description: Playlist unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          200:
            description: Playlist details with preview file revisions
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Playlist unique identifier
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    name:
                      type: string
                      description: Playlist name
                      example: "Review Playlist"
                    project_id:
                      type: string
                      format: uuid
                      description: Project unique identifier
                      example: b35b7fb5-df86-5776-b181-68564193d36
                    shots:
                      type: array
                      description: List of shots with preview file revisions
                      items:
                        type: object
                        example: [{"id": "uuid", "preview_file_id": "uuid"}]
        """
        user_service.block_access_to_vendor()
        user_service.check_project_access(project_id)
        return playlists_service.get_playlist_with_preview_file_revisions(
            playlist_id
        )


class EntityPreviewsResource(Resource):

    @jwt_required()
    def get(self, entity_id):
        """
        Get entity previews
        ---
        description: Retrieve all previews related to a given entity. It sends
          them as a dict. Keys are related task type ids and values are arrays
          of preview for this task type.
        tags:
          - Playlists
        parameters:
          - in: path
            name: entity_id
            required: true
            schema:
              type: string
              format: uuid
            description: Entity unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: All previews related to given entity grouped by task type
            content:
              application/json:
                schema:
                  type: object
                  additionalProperties:
                    type: array
                    items:
                      type: object
                      properties:
                        id:
                          type: string
                          format: uuid
                          description: Preview file unique identifier
                          example: a24a6ea4-ce75-4665-a070-57453082c25
                        name:
                          type: string
                          description: Preview file name
                          example: "preview_v001.png"
        """
        user_service.block_access_to_vendor()
        entity = entities_service.get_entity(entity_id)
        user_service.check_project_access(entity["project_id"])
        return playlists_service.get_preview_files_for_entity(entity_id)


class PlaylistAddEntityResource(Resource, ArgsMixin):

    @jwt_required()
    def post(self, playlist_id):
        """
        Add entity to playlist
        ---
        description: Atomically add a single entity to the given playlist.
        tags:
          - Playlists
        parameters:
          - in: path
            name: playlist_id
            required: true
            schema:
              type: string
              format: uuid
            description: Playlist unique identifier
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - entity_id
                properties:
                  entity_id:
                    type: string
                    format: uuid
                    description: Entity unique identifier to add to playlist
                  preview_file_id:
                    type: string
                    format: uuid
                    nullable: true
                    description: Optional preview file identifier associated to the entity
        responses:
          200:
            description: Updated playlist
            content:
              application/json:
                schema:
                  type: object
        """
        playlist = playlists_service.get_playlist(playlist_id)
        user_service.check_manager_project_access(playlist["project_id"])

        body = validation.validate_request_body(AddEntityToPlaylistSchema)
        updated_playlist = playlists_service.add_entity_to_playlist(
            playlist_id,
            str(body.entity_id),
            str(body.preview_file_id) if body.preview_file_id else None,
        )
        return updated_playlist


class PlaylistDownloadResource(Resource):

    @jwt_required()
    def get(self, playlist_id, build_job_id):
        """
        Download playlist build
        ---
        description: Download given playlist build as MP4 file.
        tags:
          - Playlists
        parameters:
          - in: path
            name: playlist_id
            required: true
            schema:
              type: string
              format: uuid
            description: Playlist unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: build_job_id
            required: true
            schema:
              type: string
              format: uuid
            description: Build job unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          200:
            description: Playlist build downloaded as MP4 file
            content:
              video/mp4:
                schema:
                  type: string
                  format: binary
          400:
            description: Build not finished, need to retry later
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    error:
                      type: boolean
                      description: Error flag
                      example: true
                    message:
                      type: string
                      description: Error message
                      example: "Build is not finished"
        """
        user_service.block_access_to_vendor()
        playlist = playlists_service.get_playlist(playlist_id)
        user_service.check_playlist_access(playlist, supervisor_access=True)
        build_job = playlists_service.get_build_job(build_job_id)
        if str(build_job["playlist_id"]) != str(playlist_id):
            abort(404)

        if build_job["status"] != "succeeded":
            return {"error": True, "message": "Build is not finished"}, 400

        project = projects_service.get_project(playlist["project_id"])
        movie_file_path = fs.get_file_path_and_file(
            config,
            file_store.get_local_movie_path,
            file_store.open_movie,
            "playlists",
            build_job_id,
            "mp4",
        )
        context_name = playlists_service.get_playlist_download_context_name(
            project, playlist
        )
        download_name = "%s_%s_%s.mp4" % (
            slugify.slugify(build_job["created_at"], separator="").replace(
                "t", "_"
            ),
            context_name,
            slugify.slugify(playlist["name"], separator="_"),
        )
        return flask_send_file(
            movie_file_path,
            conditional=True,
            mimetype="video/mp4",
            as_attachment=True,
            download_name=download_name,
        )


class BuildPlaylistMovieResource(Resource, ArgsMixin):

    @jwt_required()
    def get(self, playlist_id):
        """
        Build playlist movie
        ---
        description: Build given playlist as MP4 movie. Starts a build job that
          processes the playlist shots into a video file.
        tags:
          - Playlists
        parameters:
          - in: path
            name: playlist_id
            required: true
            schema:
              type: string
              format: uuid
            description: Playlist unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: full
            required: false
            schema:
              type: boolean
            description: Whether to build full quality movie
            example: true
        responses:
          200:
            description: Build job created for playlist movie
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Build job unique identifier
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    status:
                      type: string
                      description: Build job status
                      example: "pending"
                    created_at:
                      type: string
                      format: date-time
                      description: Build job creation timestamp
                      example: "2022-07-12T10:30:00Z"
        """
        playlist = playlists_service.get_playlist(playlist_id)
        user_service.check_manager_project_access(playlist["project_id"])

        project = projects_service.get_project(playlist["project_id"])
        (width, height) = preview_files_service.get_preview_file_dimensions(
            project
        )
        fps = preview_files_service.get_preview_file_fps(project)
        full = self.get_bool_parameter("full")
        params = EncodingParameters(width=width, height=height, fps=fps)

        shots = [
            {"preview_file_id": x.get("preview_file_id")}
            for x in playlist["shots"]
        ]

        job = playlists_service.start_build_job(playlist)
        if config.ENABLE_JOB_QUEUE:
            remote = config.ENABLE_JOB_QUEUE_REMOTE
            # remote worker can not access files local to the web app
            if remote and config.FS_BACKEND not in ["s3", "swift"]:
                return {
                    "error": True,
                    "message": "Remote job queue requires s3 or swift backend",
                }, 400

            current_user = persons_service.get_current_user()
            queue_store.job_queue.enqueue(
                playlists_service.build_playlist_job,
                args=(
                    playlist,
                    job,
                    shots,
                    params,
                    current_user["email"],
                    full,
                    remote,
                ),
                job_timeout=int(config.JOB_QUEUE_TIMEOUT),
            )
            return job
        else:
            job = playlists_service.build_playlist_movie_file(
                playlist, job, shots, params, full, remote=False
            )
            return job


class PlaylistZipDownloadResource(Resource):

    @jwt_required()
    def get(self, playlist_id):
        """
        Download playlist zip
        ---
        description: Download given playlist as ZIP file containing all preview
          files.
        tags:
          - Playlists
        parameters:
          - in: path
            name: playlist_id
            required: true
            schema:
              type: string
              format: uuid
            description: Playlist unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: Playlist downloaded as ZIP file
            content:
              application/zip:
                schema:
                  type: string
                  format: binary
        """
        user_service.block_access_to_vendor()
        playlist = playlists_service.get_playlist(playlist_id)
        user_service.check_playlist_access(playlist, supervisor_access=True)
        project = projects_service.get_project(playlist["project_id"])
        zip_file_path = playlists_service.build_playlist_zip_file(playlist)
        context_name = playlists_service.get_playlist_download_context_name(
            project, playlist
        )
        download_name = "%s_%s.zip" % (
            context_name,
            slugify.slugify(playlist["name"], separator="_"),
        )

        @after_this_request
        def cleanup(response):
            try:
                os.remove(zip_file_path)
            except OSError:
                pass
            return response

        return flask_send_file(
            zip_file_path,
            conditional=True,
            mimetype="application/zip",
            as_attachment=True,
            download_name=download_name,
        )


class BuildJobResource(Resource):

    @jwt_required()
    def get(self, playlist_id, build_job_id):
        """
        Get build job
        ---
        description: Retrieve build job related to given playlist.
        tags:
          - Playlists
        parameters:
          - in: path
            name: playlist_id
            required: true
            schema:
              type: string
              format: uuid
            description: Playlist unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: build_job_id
            required: true
            schema:
              type: string
              format: uuid
            description: Build job unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          200:
            description: Build job related to given playlist
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Build job unique identifier
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    status:
                      type: string
                      description: Build job status
                      example: "succeeded"
                    created_at:
                      type: string
                      format: date-time
                      description: Build job creation timestamp
                      example: "2022-07-12T10:30:00Z"
        """
        user_service.block_access_to_vendor()
        playlist = playlists_service.get_playlist(playlist_id)
        user_service.check_playlist_access(playlist)
        build_job = playlists_service.get_build_job(build_job_id)
        if str(build_job["playlist_id"]) != str(playlist_id):
            abort(404)
        return build_job

    @jwt_required()
    def delete(self, playlist_id, build_job_id):
        """
        Delete build job
        ---
        description: Remove given build job related to given playlist.
        tags:
          - Playlists
        parameters:
          - in: path
            name: playlist_id
            required: true
            schema:
              type: string
              format: uuid
            description: Playlist unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: build_job_id
            required: true
            schema:
              type: string
              format: uuid
            description: Build job unique identifier
            example: b35b7fb5-df86-5776-b181-68564193d36
        responses:
          204:
            description: Build job removed successfully
        """
        user_service.block_access_to_vendor()
        playlist = playlists_service.get_playlist(playlist_id)
        user_service.check_playlist_access(playlist)
        build_job = playlists_service.get_build_job(build_job_id)
        if str(build_job["playlist_id"]) != str(playlist_id):
            abort(404)
        playlists_service.remove_build_job(playlist, build_job_id)
        return "", 204


class ProjectBuildJobsResource(Resource):

    @jwt_required()
    def get(self, project_id):
        """
        Get project build jobs
        ---
        description: Retrieve all build jobs related to given project. It's
          mainly used for synchronisation purpose.
        tags:
          - Playlists
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: All build jobs related to given project
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Build job unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      status:
                        type: string
                        description: Build job status
                        example: "succeeded"
                      created_at:
                        type: string
                        format: date-time
                        description: Build job creation timestamp
                        example: "2022-07-12T10:30:00Z"
        """
        permissions.check_admin_permissions()
        projects_service.get_project(project_id)
        return playlists_service.get_build_jobs_for_project(project_id)


class ProjectAllPlaylistsResource(Resource, ArgsMixin):

    @jwt_required()
    def get(self, project_id):
        """
        Get all project playlists
        ---
        description: Retrieve all playlists related to given project. It's
          mainly used for synchronisation purpose.
        tags:
          - Playlists
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: All playlists related to given project
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Playlist unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        description: Playlist name
                        example: "Review Playlist"
                      project_id:
                        type: string
                        format: uuid
                        description: Project unique identifier
                        example: b35b7fb5-df86-5776-b181-68564193d36
        """
        permissions.check_admin_permissions()
        projects_service.get_project(project_id)
        page = self.get_page()
        return playlists_service.get_playlists_for_project(project_id, page)


class TempPlaylistResource(Resource, ArgsMixin):

    @jwt_required()
    def post(self, project_id):
        """
        Generate temp playlist
        ---
        description: Generate a temporary playlist from task IDs. It's mainly
          used for synchronisation purpose.
        tags:
          - Playlists
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            description: Project unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: sort
            required: false
            schema:
              type: boolean
            description: Whether to sort the playlist
            example: true
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - task_ids
                properties:
                  task_ids:
                    type: array
                    items:
                      type: string
                      format: uuid
                    description: List of task unique identifiers
                    example: ["a24a6ea4-ce75-4665-a070-57453082c25"]
        responses:
          200:
            description: Temporary playlist generated
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Preview file unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        description: Preview file name
                        example: "preview_v001.png"
          400:
            description: Invalid task IDs
        """
        user_service.block_access_to_vendor()
        user_service.check_project_access(project_id)
        body = validation.validate_request_body(TempPlaylistCreateSchema)
        task_ids = [str(t) for t in body.task_ids]
        sort = self.get_bool_parameter("sort")
        return (
            playlists_service.generate_temp_playlist(task_ids, sort=sort) or []
        )


class NotifyClientsResource(Resource, ArgsMixin):

    @jwt_required()
    def post(self, playlist_id):
        """
        Notify clients playlist ready
        ---
        description: Notify clients that given playlist is ready for review.
        tags:
          - Playlists
        parameters:
          - in: path
            name: playlist_id
            required: true
            schema:
              type: string
              format: uuid
            description: Playlist unique identifier
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: false
          content:
            application/json:
              schema:
                type: object
                properties:
                  studio_id:
                    type: string
                    format: uuid
                    description: Studio unique identifier to notify
                    example: b35b7fb5-df86-5776-b181-68564193d36
                  department_id:
                    type: string
                    format: uuid
                    description: Department unique identifier to notify
                    example: c46c8gc6-eg97-6887-c292-79675204e47
        responses:
          200:
            description: Clients notified successfully
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    status:
                      type: string
                      description: Notification status
                      example: "success"
        """
        playlist = playlists_service.get_playlist(playlist_id)
        user_service.check_manager_project_access(playlist["project_id"])
        body = validation.validate_request_body(
            NotifyClientsPlaylistSchema,
            data=request.get_json(silent=True) or {},
        )
        studio_id = str(body.studio_id) if body.studio_id else None
        department_id = str(body.department_id) if body.department_id else None
        notifications_service.notify_clients_playlist_ready(
            playlist, studio_id, department_id
        )
        return {"status": "success"}



AssetSchema = {
    "type": "object",
    "properties": {
        "name": {"type": "string", "description": "Name of asset"},
        "code": {
            "type": "string",
            "description": "Utility field for the pipeline to identify the asset",
        },
        "description": {
            "type": "string",
            "description": "Asset brief",
        },
        "canceled": {
            "type": "boolean",
            "default": "False",
            "description": "True if the asset has been delete one time, False otherwise",
        },
        "project_id": {
            "type": "string",
            "format": "UUID",
            "description": "Project ID",
        },
        "entity_type_id": {
            "type": "string",
            "format": "UUID",
            "description": "Asset type ID",
        },
        "source_id": {
            "type": "string",
            "format": "UUID",
            "description": "Field used to set the episode_id",
        },
        "preview_file_id": {
            "type": "string",
            "format": "UUID",
            "description": "ID of preview file used as thumbnail",
        },
        "data": {
            "type": "string",
            "format": "json",
            "description": "Free JSON field to add metadata",
        },
        "shotgun_id": {
            "type": "integer",
            "description": "Used for synchronization with a Shotgun instance",
        },
    },
}

AssetInstanceSchema = {
    "type": "object",
    "properties": {
        "asset_id": {
            "type": "string",
            "format": "UUID",
            "description": "Instantiated asset",
        },
        "name": {"type": "string"},
        "number": {"type": "integer"},
        "description": {"type": "string"},
        "active": {"type": "boolean", "default": "True"},
        "data": {
            "type": "string",
            "format": "json",
            "description": "Free JSON field to add metadata",
        },
        "scene_id": {
            "type": "string",
            "format": "UUID",
            "description": "Target scene",
        },
        "target_asset_id": {
            "type": "string",
            "format": "UUID",
            "description": "Use when instantiating an asset in an asset is required",
        },
    },
}

AssetTypeSchema = {
    "type": "object",
    "properties": {"name": {"type": "string"}},
}

AttachmentFileSchema = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "description": "Name of attachment file",
        },
        "size": {
            "type": "integer",
            "description": "Size of attachment file",
        },
        "extension": {
            "type": "string",
            "description": "Extension of attachment file",
        },
        "mimetype": {"type": "string"},
        "comment_id": {
            "type": "string",
            "format": "UUID",
            "description": "Comment to which the file is attached",
        },
    },
}

BuildJobSchema = {
    "type": "object",
    "properties": {
        "status": {
            "type": "string",
            "description": "Status of build job (running, failed, succeeded)",
        },
        "job_type": {
            "type": "string",
            "description": "Type of build job (archive, movie)",
        },
        "ended_at": {"type": "string", "format": "date-time"},
        "playlist_id": {
            "type": "string",
            "format": "UUID",
            "description": "Playlist ID",
        },
    },
}

CommentSchema = {
    "type": "object",
    "properties": {
        "shotgun_id": {
            "type": "integer",
            "description": "Used for synchronization with a Shotgun instance",
        },
        "object_id": {
            "type": "string",
            "format": "UUID",
            "description": "Unique ID of the commented model instance",
        },
        "object_type": {
            "type": "string",
            "description": "Model type of the comment model instance",
        },
        "text": {"type": "string"},
        "data": {
            "type": "string",
            "format": "json",
            "description": "Free JSON field to add metadata",
        },
        "replies": {
            "type": "string",
            "format": "json",
            "default": "[]",
        },
        "checklist": {"type": "string", "format": "json"},
        "pinned": {"type": "boolean"},
        "task_status_id": {
            "type": "string",
            "format": "UUID",
            "description": "Task status attached to comment",
        },
        "person_id": {
            "type": "string",
            "format": "UUID",
            "description": "The person who publishes the comment",
        },
        "preview_file_id": {
            "type": "string",
            "format": "UUID",
            "description": "ID of preview file used as thumbnail",
        },
    },
}

CustomActionSchema = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "description": "Name of custom action",
        },
        "url": {"type": "string"},
        "entity_type": {"type": "string", "default": "all"},
        "is_ajax": {
            "type": "boolean",
            "default": "False",
            "description": "True if the custom action is ajax, False otherwise",
        },
    },
}

DataImportErrorSchema = {
    "type": "object",
    "properties": {
        "event_data": {
            "type": "string",
            "format": "json",
            "description": "JSON field to add event data",
        },
        "source": {
            "type": "array",
            "items": {"type": "string", "enum": ["csv", "shotgun"]},
        },
    },
}

DayOffSchema = {
    "type": "object",
    "properties": {
        "date": {"type": "string", "format": "date"},
        "person_id": {
            "type": "string",
            "format": "UUID",
            "description": "The person who will take the day off",
        },
    },
}

DepartmentSchema = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "description": "Name of department",
        },
        "color": {
            "type": "string",
            "description": "Color of department",
        },
        "archived": {
            "type": "boolean",
            "description": "True if the department is archived else False",
        },
    },
}

DesktopLoginLogSchema = {
    "type": "object",
    "properties": {
        "date": {"type": "string", "format": "date"},
        "person_id": {
            "type": "string",
            "format": "UUID",
            "description": "Person ID",
        },
    },
}

EpisodeSchema = {
    "type": "object",
    "properties": {
        "name": {"type": "string", "description": "Name of episode"},
        "code": {
            "type": "string",
            "description": "Utility field for the pipeline to identify the episode",
        },
        "description": {
            "type": "string",
            "description": "Episode brief",
        },
        "canceled": {
            "type": "boolean",
            "default": "False",
            "description": "True if the episode has been delete one time, False otherwise",
        },
        "project_id": {
            "type": "string",
            "format": "UUID",
            "description": "Project ID",
        },
        "source_id": {
            "type": "string",
            "format": "UUID",
            "description": "Field used to set the episode_id",
        },
        "preview_file_id": {
            "type": "string",
            "format": "UUID",
            "description": "ID of preview file used as thumbnail",
        },
        "data": {
            "type": "string",
            "format": "json",
            "description": "Free JSON field to add metadata",
        },
        "shotgun_id": {
            "type": "integer",
            "description": "Used for synchronization with a Shotgun instance",
        },
    },
}

EventSchema = {
    "type": "object",
    "properties": {
        "name": {"type": "string", "description": "Name of event"},
        "user_id": {
            "type": "string",
            "format": "UUID",
            "description": "The user who made the action that emitted the event",
        },
        "project_id": {
            "type": "string",
            "format": "UUID",
            "description": "Project ID",
        },
        "data": {
            "type": "string",
            "format": "json",
            "description": "Free JSON field to add metadata",
        },
    },
}

FileStatusSchema = {
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "color": {"type": "string"},
    },
}

LoginLogSchema = {
    "type": "object",
    "properties": {
        "origin": {"type": "string", "description": "web, script"},
        "ip_address": {
            "type": "string",
            "description": "IP address of device used to login",
        },
        "person_id": {
            "type": "string",
            "format": "UUID",
            "description": "Person ID",
        },
    },
}

MetadataSchema = {
    "type": "object",
    "properties": {
        "project_id": {
            "type": "string",
            "format": "UUID",
            "description": "ID of project for which metadata are added",
        },
        "entity_type": {
            "type": "string",
            "description": "Asset or Shot",
        },
        "name": {
            "type": "string",
            "description": "Field name for GUI",
        },
        "field_name": {
            "type": "string",
            "description": "Technical field name",
        },
        "choices": {
            "type": "string",
            "format": "json",
            "description": "Array of string that represents the available values for this metadata (this metadata is considered as a free field if this array is empty)",
        },
        "for_client": {"type": "boolean"},
    },
}

MilestoneSchema = {
    "type": "object",
    "properties": {
        "data": {
            "type": "string",
            "format": "date",
            "description": "Milestone date of production schedule",
        },
        "name": {"type": "string", "description": "Name of milestone"},
        "task_type_id": {
            "type": "string",
            "format": "UUID",
            "description": "Task type ID",
        },
        "project_id": {
            "type": "string",
            "format": "UUID",
            "description": "Project ID",
        },
    },
}

NewsSchema = {
    "type": "object",
    "properties": {
        "change": {"type": "boolean", "default": "False"},
        "author_id": {
            "type": "string",
            "format": "UUID",
            "description": "Person who wrote the comment",
        },
        "comment_id": {
            "type": "string",
            "format": "UUID",
            "description": "Posted comment ID",
        },
        "task_id": {
            "type": "string",
            "format": "UUID",
            "description": "Task ID",
        },
        "preview_file_id": {
            "type": "string",
            "format": "UUID",
            "description": "Preview file ID",
        },
    },
}

NotificationSchema = {
    "type": "object",
    "properties": {
        "read": {
            "type": "boolean",
            "description": "True if user read it, False otherwise",
        },
        "change": {
            "type": "boolean",
            "description": "True if there is status change related to this status, False otherwise",
        },
        "person_id": {
            "type": "string",
            "format": "UUID",
            "description": "The user to who the notification is aimed at",
        },
        "author_id": {
            "type": "string",
            "format": "UUID",
            "description": "Author of the event to notify",
        },
        "comment_id": {
            "type": "string",
            "format": "UUID",
            "description": "Comment related to the notification, if there is any",
        },
        "task_id": {
            "type": "string",
            "format": "UUID",
            "description": "Task related to the notification, if there is any",
        },
        "reply_id": {
            "type": "string",
            "format": "UUID",
            "description": "Reply related to notification",
        },
        "type": {
            "type": "string",
            "description": "Type of notification",
        },
    },
}

OrganisationSchema = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "description": "Name of organisation",
        },
        "hours_by_day": {"type": "float"},
        "has_avatar": {
            "type": "boolean",
            "default": "False",
            "description": "True if the organisation has an avatar, False otherwise",
        },
        "use_original_file_name": {
            "type": "boolean",
            "default": "False",
            "description": "True if the organisation uses original file names, False otherwise",
        },
        "timesheets_locked": {
            "type": "boolean",
            "default": "False",
            "description": "True if the organisation's timesheets are locked, False otherwise",
        },
        "hd_by_default": {"type": "boolean", "default": "False"},
        "chat_token_slack": {"type": "string"},
        "chat_webhook_mattermost": {"type": "string"},
        "chat_token_discord": {"type": "string"},
    },
}

OutputFileSchema = {
    "type": "object",
    "properties": {
        "shotgun_id": {
            "type": "integer",
            "description": "Used for synchronization with a Shotgun instance",
        },
        "name": {
            "type": "string",
            "description": "Name of output file",
        },
        "extension": {
            "type": "string",
            "description": "Extension of output file",
        },
        "description": {
            "type": "string",
            "description": "Output file brief",
        },
        "comment": {
            "type": "string",
            "description": "Comment on output file",
        },
        "revision": {
            "type": "integer",
            "description": "Revision number of output file",
        },
        "size": {
            "type": "integer",
            "description": "Size of output file",
        },
        "checksum": {
            "type": "string",
            "description": "Checksum of output file",
        },
        "source": {
            "type": "string",
            "description": "Created by a script, a webgui or a desktop gui",
        },
        "path": {
            "type": "string",
            "description": "File path on the production hard drive",
        },
        "representation": {
            "type": "string",
            "description": "Precise what kind of output it is (abc, jpgs, pngs, etc.)",
        },
        "nb_elements": {
            "type": "integer",
            "default": "1",
            "description": "For image sequence",
        },
        "canceled": {
            "type": "boolean",
        },
        "file_status_id": {
            "type": "string",
            "format": "UUID",
            "description": "File status ID",
        },
        "entity_id": {
            "type": "string",
            "format": "UUID",
            "description": "Asset or Shot concerned by the output file",
        },
        "asset_instance_id": {
            "type": "string",
            "format": "UUID",
            "description": "Asset instance ID",
        },
        "output_type_id": {
            "type": "string",
            "format": "UUID",
            "description": "Type of output (geometry, cache, etc.)",
        },
        "task_type_id": {
            "type": "string",
            "format": "UUID",
            "description": "Task type related to this output file (modeling, animation, etc.)",
        },
        "person_id": {
            "type": "string",
            "format": "UUID",
            "description": "Author of the file",
        },
        "source_file_id": {
            "type": "string",
            "format": "UUID",
            "description": "Working file that led to create this output file",
        },
        "temporal_entity_id": {
            "type": "string",
            "format": "UUID",
            "description": "Shot, scene or sequence needed for output files related to an asset instance",
        },
        "data": {
            "type": "string",
            "format": "json",
            "description": "Free JSON field to add metadata",
        },
    },
}

OutputTypeSchema = {
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "short_name": {"type": "string"},
    },
}

PersonSchema = {
    "type": "object",
    "properties": {
        "first_name": {"type": "string"},
        "last_name": {"type": "string"},
        "email": {"type": "string", "description": "Serve as login"},
        "phone": {"type": "string"},
        "active": {
            "type": "boolean",
            "description": "True if the person is still in the studio, False otherwise",
        },
        "last_presence": {
            "type": "string",
            "format": "date",
            "description": "Last time the person worked for the studio",
        },
        "password": {"type": "string", "format": "byte"},
        "desktop_login": {
            "type": "string",
            "description": "Login used on desktop",
        },
        "shotgun_id": {
            "type": "integer",
            "description": "Used for synchronization with a Shotgun instance",
        },
        "timezone": {"type": "string"},
        "locale": {"type": "string"},
        "data": {
            "type": "string",
            "format": "json",
            "description": "Free JSON field to add metadata",
        },
        "role": {"type": "string", "default": "user"},
        "has_avatar": {
            "type": "boolean",
            "default": "False",
            "description": "True if user has an avatar, False otherwise",
        },
        "notifications_enabled": {
            "type": "boolean",
        },
        "notifications_slack_enabled": {
            "type": "boolean",
        },
        "notifications_slack_userid": {
            "type": "string",
        },
        "notifications_mattermost_enabled": {
            "type": "boolean",
        },
        "notifications_mattermost_userid": {
            "type": "string",
        },
        "notifications_discord_enabled": {
            "type": "boolean",
        },
        "notifications_discord_userid": {
            "type": "string",
        },
    },
}

PlaylistSchema = {
    "type": "object",
    "properties": {
        "name": {"type": "string", "description": "Name of playlist"},
        "shots": {
            "type": "string",
            "format": "json",
            "description": "JSON field describing shot and preview listed in",
        },
        "project_id": {
            "type": "string",
            "format": "UUID",
            "description": "Project ID",
        },
        "episode_id": {
            "type": "string",
            "format": "UUID",
            "description": "Episode ID",
        },
        "for_client": {"type": "boolean", "default": "False"},
        "for_entity": {"type": "string", "default": "shot"},
        "is_for_all": {"type": "boolean", "default": "False"},
        "task_type_id": {
            "type": "string",
            "format": "UUID",
            "description": "Task type ID",
        },
    },
}

PreviewFileSchema = {
    "type": "object",
    "properties": {
        "shotgun_id": {
            "type": "integer",
            "description": "Used for synchronization with a Shotgun instance",
        },
        "name": {
            "type": "string",
            "description": "Name of preview file",
        },
        "original_name": {
            "type": "string",
            "description": "Original name of preview file",
        },
        "revision": {
            "type": "integer",
            "default": "1",
            "description": "Revision number of preview file",
        },
        "position": {
            "type": "integer",
            "default": "1",
            "description": "Position of preview file",
        },
        "extension": {
            "type": "string",
            "description": "Extension of preview file",
        },
        "description": {
            "type": "string",
            "description": "Preview file brief",
        },
        "path": {
            "type": "string",
            "description": "File path on the production hard drive",
        },
        "source": {
            "type": "string",
            "description": "Created by a script, a webgui or a desktop gui",
        },
        "file_size": {
            "type": "integer",
            "default": "0",
            "description": "Size of output file",
        },
        "comment": {
            "type": "string",
            "description": "Comment on output file",
        },
        "checksum": {
            "type": "string",
            "description": "Checksum of output file",
        },
        "representation": {
            "type": "string",
            "description": "Precise what kind of output it is (abc, jpgs, pngs, etc.)",
        },
        "nb_elements": {
            "type": "integer",
            "default": "1",
            "description": "For image sequence",
        },
        "canceled": {
            "type": "boolean",
        },
        "file_status_id": {
            "type": "string",
            "format": "UUID",
            "description": "File status ID",
        },
        "entity_id": {
            "type": "string",
            "format": "UUID",
            "description": "Asset or Shot concerned by the output file",
        },
        "asset_instance_id": {
            "type": "string",
            "format": "UUID",
            "description": "Asset instance ID",
        },
        "output_type_id": {
            "type": "string",
            "format": "UUID",
            "description": "Type of output (geometry, cache, etc.)",
        },
        "task_type_id": {
            "type": "string",
            "format": "UUID",
            "description": "Task type related to this output file (modeling, animation, etc.)",
        },
        "person_id": {
            "type": "string",
            "format": "UUID",
            "description": "Author of the file",
        },
        "source_file_id": {
            "type": "string",
            "format": "UUID",
            "description": "Working file that led to create this output file",
        },
        "temporal_entity_id": {
            "type": "string",
            "format": "UUID",
            "description": "Shot, scene or sequence needed for output files related to an asset instance",
        },
        "data": {
            "type": "string",
            "format": "json",
            "description": "Free JSON field to add metadata",
        },
    },
}

ProjectSchema = {
    "type": "object",
    "properties": {
        "name": {"type": "string", "description": "Name of project"},
        "code": {
            "type": "string",
            "description": "Utility field for the pipeline to identify the project",
        },
        "description": {
            "type": "string",
            "description": "Project brief",
        },
        "shotgun_id": {
            "type": "integer",
            "description": "Used for synchronization with a Shotgun instance",
        },
        "file_tree": {
            "type": "string",
            "format": "json",
            "description": "Templates to use to build file paths",
        },
        "data": {
            "type": "string",
            "format": "json",
            "description": "Free JSON field to add metadata",
        },
        "project_status_id": {
            "type": "string",
            "format": "UUID",
            "description": "Project status ID",
        },
        "has_avatar": {
            "type": "boolean",
            "default": "False",
            "description": "True if the project has an avatar, False otherwise",
        },
        "fps": {"type": "string", "description": "Frames per second"},
        "ratio": {"type": "string"},
        "resolution": {"type": "string"},
        "production_type": {
            "type": "string",
            "description": "short, featurefilm or tvshow",
            "default": "short",
        },
        "end_date": {"type": "string", "format": "date"},
        "start_date": {"type": "string", "format": "date"},
        "man_days": {
            "type": "integer",
            "description": "Estimated number of working days to complete the project",
        },
        "nb_episodes": {"type": "integer", "default": "0"},
        "episode_span": {"type": "integer", "default": "0"},
        "max_retakes": {"type": "integer", "default": "0"},
        "is_clients_isolated": {
            "type": "boolean",
            "default": "False",
            "description": "True if the clients are isolated from the project, False otherwise",
        },
    },
}

ProjectStatusSchema = {
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "color": {"type": "string"},
    },
}

ScheduleItemSchema = {
    "type": "object",
    "properties": {
        "start_date": {"type": "string", "format": "date"},
        "end_date": {"type": "string", "format": "date"},
        "man_days": {"type": "integer"},
        "project_id": {
            "type": "string",
            "format": "UUID",
            "description": "Project ID",
        },
        "task_type_id": {
            "type": "string",
            "format": "UUID",
            "description": "Task type ID",
        },
        "object_id": {
            "type": "string",
            "format": "UUID",
            "description": "Object ID",
        },
    },
}

SearchFilterSchema = {
    "type": "object",
    "properties": {
        "list_type": {"type": "string", "description": "Type of list"},
        "entity_type": {
            "type": "string",
            "description": "Type of entity",
        },
        "name": {
            "type": "string",
            "description": "Name of search filter",
        },
        "search_query": {"type": "string"},
        "person_id": {
            "type": "string",
            "format": "UUID",
            "description": "Person ID",
        },
        "project_id": {
            "type": "string",
            "format": "UUID",
            "description": "Project ID",
        },
    },
}

SequenceSchema = {
    "type": "object",
    "properties": {
        "name": {"type": "string", "description": "Name of sequence"},
        "code": {
            "type": "string",
            "description": "Utility field for the pipeline to identify the sequence",
        },
        "description": {
            "type": "string",
            "description": "Sequence brief",
        },
        "canceled": {
            "type": "boolean",
            "default": "False",
            "description": "True if the sequence has been delete one time, False otherwise",
        },
        "project_id": {
            "type": "string",
            "format": "UUID",
            "description": "Project ID",
        },
        "parent_id": {
            "type": "string",
            "format": "UUID",
            "description": "Episode ID",
        },
        "source_id": {
            "type": "string",
            "format": "UUID",
            "description": "Field used to set the episode_id",
        },
        "preview_file_id": {
            "type": "string",
            "format": "UUID",
            "description": "ID of preview file used as thumbnail",
        },
        "data": {
            "type": "string",
            "format": "json",
            "description": "Free JSON field to add metadata",
        },
        "shotgun_id": {
            "type": "integer",
            "description": "Used for synchronization with a Shotgun instance",
        },
    },
}

ShotSchema = {
    "type": "object",
    "properties": {
        "name": {"type": "string", "description": "Name of shot"},
        "code": {
            "type": "string",
            "description": "Utility field for the pipeline to identify the shot",
        },
        "description": {"type": "string", "description": "Shot brief"},
        "canceled": {
            "type": "boolean",
            "default": "False",
            "description": "True if the shot has been delete one time, False otherwise",
        },
        "project_id": {
            "type": "string",
            "format": "UUID",
            "description": "Project ID",
        },
        "parent_id": {
            "type": "string",
            "format": "UUID",
            "description": "Episode ID",
        },
        "entity_type_id": {
            "type": "string",
            "format": "UUID",
            "description": "Shot type ID",
        },
        "source_id": {
            "type": "string",
            "format": "UUID",
            "description": "Field used to set the episode_id",
        },
        "preview_file_id": {
            "type": "string",
            "format": "UUID",
            "description": "ID of preview file used as thumbnail",
        },
        "data": {
            "type": "string",
            "format": "json",
            "description": "Free JSON field to add metadata",
        },
        "shotgun_id": {
            "type": "integer",
            "description": "Used for synchronization with a Shotgun instance",
        },
    },
}

SoftwareSchema = {
    "type": "object",
    "properties": {
        "name": {"type": "string", "description": "Name of software"},
        "short_name": {
            "type": "string",
            "description": "Short name of software",
        },
        "file_extension": {
            "type": "string",
            "description": "Main extension used for this software's files",
        },
        "secondary_extensions": {
            "type": "string",
            "format": "json",
            "description": "Other extensions used for this software's files",
        },
    },
}

StatusAutomationSchema = {
    "type": "object",
    "properties": {
        "entity_type": {"type": "string", "default": "asset"},
        "in_task_type_id": {
            "type": "string",
            "format": "UUID",
            "description": "Task type ID",
        },
        "in_task_status_id": {
            "type": "string",
            "format": "UUID",
            "description": "Task status ID",
        },
        "out_field_type": {
            "type": "string",
            "description": "Field type (status, ready_for)",
        },
        "out_task_type_id": {
            "type": "string",
            "format": "UUID",
            "description": "Task type ID",
        },
        "out_task_status_id": {
            "type": "string",
            "format": "UUID",
            "description": "Task status ID",
        },
    },
}

SubscriptionToNotificationsSchema = {
    "type": "object",
    "properties": {
        "person_id": {
            "type": "string",
            "format": "UUID",
            "description": "Person ID",
        },
        "task_id": {
            "type": "string",
            "format": "UUID",
            "description": "Task ID",
        },
        "entity_id": {
            "type": "string",
            "format": "UUID",
            "description": "Entity ID",
        },
        "task_type_id": {
            "type": "string",
            "format": "UUID",
            "description": "Task type ID",
        },
    },
}

StudioSchema = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "description": "Name of studio",
        },
        "color": {
            "type": "string",
            "description": "Color of studio",
        },
        "archived": {
            "type": "boolean",
            "description": "True if the studio is archived else False",
        },
    },
}

TaskSchema = {
    "type": "object",
    "properties": {
        "name": {"type": "string", "description": "Name of task"},
        "description": {"type": "string", "description": "Task brief"},
        "priority": {
            "type": "integer",
            "default": "0",
            "description": "Priority of task",
        },
        "duration": {
            "type": "float",
            "default": "0",
            "description": "Duration of task",
        },
        "estimation": {
            "type": "float",
            "default": "0",
            "description": "Estimation of duration of task",
        },
        "completion_rate": {
            "type": "integer",
            "default": "0",
            "description": "Completion rate of task",
        },
        "retake_count": {
            "type": "integer",
            "default": "0",
            "description": "Retake count of task",
        },
        "sort_order": {
            "type": "integer",
            "default": "0",
            "description": "Sort order of task",
        },
        "start_date": {"type": "string", "format": "date-time"},
        "due_date": {"type": "string", "format": "date-time"},
        "real_start_date": {"type": "string", "format": "date-time"},
        "end_date": {"type": "string", "format": "date-time"},
        "last_comment_date": {"type": "string", "format": "date-time"},
        "nb_assets_ready": {
            "type": "integer",
            "default": "0",
            "description": "Number of assets ready",
        },
        "data": {
            "type": "string",
            "format": "json",
            "description": "Free JSON field to add metadata",
        },
        "shotgun_id": {
            "type": "integer",
            "description": "Used for synchronization with a Shotgun instance",
        },
        "project_id": {
            "type": "string",
            "format": "UUID",
            "description": "Project ID",
        },
        "task_type_id": {
            "type": "string",
            "format": "UUID",
            "description": "Task type ID",
        },
        "task_status_id": {
            "type": "string",
            "format": "UUID",
            "description": "Task status ID",
        },
        "entity_id": {
            "type": "string",
            "format": "UUID",
            "description": "Entity ID",
        },
        "assigner_id": {
            "type": "string",
            "format": "UUID",
            "description": "Person ID",
        },
    },
}

TaskStatusSchema = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "description": "Name of task status",
        },
        "short_name": {
            "type": "string",
            "description": "Short name of task status",
        },
        "color": {"type": "string"},
        "is_done": {
            "type": "boolean",
            "default": "False",
            "description": "True if the task is done, False otherwise",
        },
        "is_artist_allowed": {
            "type": "boolean",
            "default": "True",
            "description": "True if the artist is allowed, False otherwise",
        },
        "is_client_allowed": {
            "type": "boolean",
            "default": "True",
            "description": "True if the client is allowed, False otherwise",
        },
        "is_retake": {
            "type": "boolean",
            "default": "False",
            "description": "True if the task was retaken, False otherwise",
        },
        "is_feedback_request": {
            "type": "boolean",
            "default": "False",
            "description": "True if feedback was requested, False otherwise",
        },
        "is_default": {
            "type": "boolean",
            "default": "False",
            "description": "True if the task is default, False otherwise",
        },
        "shotgun_id": {
            "type": "integer",
            "description": "Used for synchronization with a Shotgun instance",
        },
    },
}

TaskTypeSchema = {
    "type": "object",
    "properties": {
        "name": {"type": "string", "description": "Name of task type"},
        "short_name": {
            "type": "string",
            "description": "Short name of task type",
        },
        "color": {"type": "string", "default": "#FFFFFF"},
        "priority": {
            "type": "integer",
            "default": "1",
            "description": "Priority of task type",
        },
        "for_entity": {"type": "string", "default": "Asset"},
        "allow_timelog": {
            "type": "boolean",
            "default": "True",
            "description": "True if timelog is allowed, False otherwise",
        },
        "shotgun_id": {
            "type": "integer",
            "description": "Used for synchronization with a Shotgun instance",
        },
        "department_id": {
            "type": "string",
            "format": "UUID",
            "description": "Department ID",
        },
    },
}

TimeSpentSchema = {
    "type": "object",
    "properties": {
        "duration": {"type": "float"},
        "date": {"type": "string", "format": "date"},
        "task_id": {
            "type": "string",
            "format": "UUID",
            "description": "Related task ID",
        },
        "person_id": {
            "type": "string",
            "format": "UUID",
            "description": "The person who performed the working time",
        },
    },
}

WorkingFileSchema = {
    "type": "object",
    "properties": {
        "shotgun_id": {
            "type": "integer",
            "description": "Used for synchronization with a Shotgun instance",
        },
        "name": {
            "type": "string",
            "description": "Name of working file",
        },
        "description": {
            "type": "string",
            "description": "working file brief",
        },
        "comment": {
            "type": "string",
            "description": "Comment on working file",
        },
        "revision": {
            "type": "integer",
            "description": "Revision number of working file",
        },
        "size": {
            "type": "integer",
            "description": "Size of working file",
        },
        "checksum": {
            "type": "string",
            "description": "Checksum of working file",
        },
        "path": {
            "type": "string",
            "description": "File path on the production hard drive",
        },
        "data": {
            "type": "string",
            "format": "json",
            "description": "Free JSON field to add metadata",
        },
        "task_id": {
            "type": "string",
            "format": "UUID",
            "description": "Task for which the working file is made",
        },
        "entity_id": {
            "type": "string",
            "format": "UUID",
            "description": "Entity for which the working is made",
        },
        "person_id": {
            "type": "string",
            "format": "UUID",
            "description": "Author of the file",
        },
        "software_id": {
            "type": "string",
            "format": "UUID",
            "description": "Software used to build this working file",
        },
    },
}



from flask_restful import Resource
from flask_jwt_extended import jwt_required

from zou.app.mixin import ArgsMixin
from zou.app.utils import fields, permissions, date_helpers

from zou.app.services import events_service
from zou.app.services.exception import WrongParameterException


class EventsResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self):
        """
        Get events
        ---
        description: Retrieve last events with filtering support. Filters can be
          specified in the query string to narrow down results by date range,
          project, or other criteria.
        tags:
          - Events
        parameters:
          - in: query
            name: after
            type: string
            format: date
            example: "2022-07-12"
            description: Filter events after this date
          - in: query
            name: before
            type: string
            format: date
            example: "2022-07-12"
            description: Filter events before this date
          - in: query
            name: only_files
            type: boolean
            default: false
            description: Return only file-related events
            example: false
          - in: query
            name: cursor_event_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: ID of the last event from previous page for cursor-based pagination
          - in: query
            name: limit
            type: integer
            default: 100
            example: 100
            description: Maximum number of events to return
          - in: query
            name: project_id
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Filter events by specific project
          - in: query
            name: name
            type: string
            example: "user_login"
            description: Filter events by event name
        responses:
          200:
            description: List of events successfully retrieved
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Event unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        description: Event name
                        example: "user_login"
                      data:
                        type: object
                        description: Event data content
                        example: {"user_id": "b35b7fb5-df86-5776-b181-68564193d36"}
                      project_id:
                        type: string
                        format: uuid
                        description: Project identifier
                        example: c46c8gc6-eg97-6887-c292-79675204e47
                      created_at:
                        type: string
                        format: date-time
                        description: Event timestamp
                        example: "2023-01-01T12:00:00Z"
                      user_id:
                        type: string
                        format: uuid
                        description: User identifier who triggered the event
                        example: d57d9hd7-fh08-7998-d403-80786315f58
        """
        args = self.get_args(
            [
                ("after", None, False),
                ("before", None, False),
                ("only_files", False, False),
                ("cursor_event_id", None, False),
                ("limit", 100, False, int),
                ("project_id", None, False),
                ("name", None, False),
            ],
        )

        permissions.check_manager_permissions()
        before = self.parse_date_parameter(args["before"])
        after = self.parse_date_parameter(args["after"])
        cursor_event_id = args["cursor_event_id"]
        limit = min(args["limit"], 1000)
        only_files = args["only_files"] == "true"
        project_id = args.get("project_id", None)
        name = args["name"]
        if project_id is not None and not fields.is_valid_id(project_id):
            raise WrongParameterException(
                "The project_id parameter is not a valid id"
            )
        if cursor_event_id is not None and not fields.is_valid_id(
            cursor_event_id
        ):
            raise WrongParameterException(
                "The cursor_event_id parameter is not a valid id"
            )
        return events_service.get_last_events(
            after=after,
            before=before,
            cursor_event_id=cursor_event_id,
            limit=limit,
            only_files=only_files,
            project_id=project_id,
            name=name,
        )


class LoginLogsResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self):
        """
        Get login logs
        ---
        description: Retrieve all login logs with filtering support. Filters can
          be specified in the query string to narrow down results by date range
          and limit.
        tags:
          - Events
        parameters:
          - in: query
            name: before
            type: string
            format: date-time
            example: "2022-07-12T00:00:00"
            description: Filter logs before this date and time
          - in: query
            name: limit
            type: integer
            example: 100
            description: Maximum number of login logs to return
        responses:
          200:
            description: List of login logs successfully retrieved
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Login log unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      user_id:
                        type: string
                        format: uuid
                        description: User identifier
                        example: b35b7fb5-df86-5776-b181-68564193d36
                      ip_address:
                        type: string
                        description: IP address of the login
                        example: "192.168.1.100"
                      user_agent:
                        type: string
                        description: User agent string
                        example: "Mozilla/5.0 (Windows NT 10.0; Win64; x64)"
                      success:
                        type: boolean
                        description: Whether the login was successful
                        example: true
                      created_at:
                        type: string
                        format: date-time
                        description: Login timestamp
                        example: "2023-01-01T12:00:00Z"
        """
        args = self.get_args(["before", ("limit", 100, False, int)])

        permissions.check_manager_permissions()
        before = None
        if args["before"] is not None:
            before = date_helpers.get_datetime_from_string(args["before"])
        limit = args["limit"]
        return events_service.get_last_login_logs(before, limit)



"""
Pydantic schemas for request/response validation in the persons blueprint.

Use these schemas to ensure incoming JSON bodies match the expected format
and to return clear validation errors (e.g. 400 with field-level messages).
"""

from datetime import datetime
from typing import Optional
from uuid import UUID

from pydantic import BaseModel, Field, field_validator


def _parse_datetime(v):
    """Parse date string (YYYY-MM-DD or ISO datetime) to datetime."""
    if v is None or isinstance(v, datetime):
        return v
    from zou.app.utils import date_helpers

    try:
        return date_helpers.get_datetime_from_string(v)
    except Exception:
        try:
            return date_helpers.get_date_from_string(v)
        except Exception:
            raise ValueError(
                "Wrong date format. Expected: 2020-01-05T13:23:10 or 2020-01-05"
            )


class DesktopLoginCreateSchema(BaseModel):
    """Body for creating a desktop login log. Date is optional (default: now)."""

    date: Optional[datetime] = None

    @field_validator("date", mode="before")
    @classmethod
    def parse_date(cls, v):
        return _parse_datetime(v) if v else v


class AddToDepartmentSchema(BaseModel):
    """Body for adding a person to a department."""

    department_id: UUID = Field(
        ..., description="Department unique identifier"
    )

    @field_validator("department_id", mode="before")
    @classmethod
    def ensure_uuid(cls, v):
        if isinstance(v, str):
            from zou.app.utils.fields import is_valid_id

            if not is_valid_id(v):
                raise ValueError("department_id must be a valid UUID")
            return v
        return v


class ChangePasswordSchema(BaseModel):
    """Body for changing a person's password (admin)."""

    password: str = Field(..., min_length=1, description="New password")
    password_2: str = Field(
        ..., min_length=1, description="Password confirmation"
    )

    @field_validator("password_2")
    @classmethod
    def passwords_match(cls, v, info):
        if "password" in info.data and info.data["password"] != v:
            raise ValueError("Confirmation password doesn't match.")
        return v



import re
import traceback
from io import StringIO
from html.parser import HTMLParser
from flask_mail import Message

from zou.app import mail, app


def send_email(subject, html, recipient_email, body=None, locale=None):
    """
    Send an email with given subject and body to given recipient.
    If locale is provided (e.g. "en_US", "fr_FR"), the Content-Language
    header is set so the recipient's client can use the correct language.
    """
    if body is None:
        body = strip_html_tags(html)
    if app.config["MAIL_DEBUG_BODY"]:
        print(body)
    if app.config["MAIL_ENABLED"]:
        with app.app_context():
            try:
                mail_default_sender = app.config["MAIL_DEFAULT_SENDER"]
                message = Message(
                    sender="Kitsu Bot <%s>" % mail_default_sender,
                    body=body,
                    html=html,
                    subject=subject,
                    recipients=[recipient_email],
                )
                if locale:
                    # Set Content-Language so the recipient's client knows the language
                    try:
                        lang_tag = locale.replace("_", "-")
                        if hasattr(message, "msg"):
                            message.msg["Content-Language"] = lang_tag
                    except Exception:
                        pass
                mail.send(message)
            except Exception:
                app.logger.info("Exception when sending a mail notification:")
                app.logger.info(traceback.format_exc())


class HTMLStripper(HTMLParser):
    def __init__(self):
        super().__init__()
        self.reset()
        self.strict = False
        self.convert_charrefs = True
        self.text = StringIO()
        self.in_link = False
        self.link_url = ""

    def handle_starttag(self, tag, attrs):
        if tag == "a":
            self.in_link = True
            for attr_name, attr_value in attrs:
                if attr_name == "href":
                    self.link_url = attr_value
                    break
        elif tag in ("p", "br", "div"):
            self.text.write("\n")

    def handle_endtag(self, tag):
        if tag == "a" and self.in_link:
            if self.link_url:
                self.text.write(" (%s)" % self.link_url)
            self.in_link = False
            self.link_url = ""
        elif tag in ("p", "div"):
            self.text.write("\n")

    def handle_data(self, d):
        self.text.write(d)

    def get_data(self):
        return self.text.getvalue()


def strip_html_tags(html):
    """
    Convert HTML email to plain text, preserving URLs from links
    and maintaining basic formatting with line breaks.
    """
    s = HTMLStripper()
    s.feed(html)
    text = s.get_data()
    # Clean up multiple consecutive newlines
    text = re.sub(r"\n{3,}", "\n\n", text)
    return text.strip()



from flask_jwt_extended import jwt_required

from zou.app.models.task_type import TaskType
from zou.app.models.schedule_item import ScheduleItem
from zou.app.services.exception import WrongParameterException
from zou.app.services import tasks_service

from zou.app.blueprints.crud.base import BaseModelResource, BaseModelsResource


class TaskTypesResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, TaskType)

    def check_read_permissions(self, options=None):
        return True

    @jwt_required()
    def get(self):
        """
        Get task types
        ---
        tags:
          - Crud
        description: Retrieve all task types. Supports filtering via
          query parameters and pagination.
        parameters:
          - in: query
            name: page
            required: false
            schema:
              type: integer
            example: 1
            description: Page number for pagination
          - in: query
            name: limit
            required: false
            schema:
              type: integer
            example: 50
            description: Number of results per page
          - in: query
            name: relations
            required: false
            schema:
              type: boolean
            default: false
            example: false
            description: Whether to include relations
        responses:
            200:
              description: Task types retrieved successfully
              content:
                application/json:
                  schema:
                    oneOf:
                      - type: array
                        items:
                          type: object
                      - type: object
                        properties:
                          data:
                            type: array
                            items:
                              type: object
                            example: []
                          total:
                            type: integer
                            example: 100
                          nb_pages:
                            type: integer
                            example: 2
                          limit:
                            type: integer
                            example: 50
                          offset:
                            type: integer
                            example: 0
                          page:
                            type: integer
                            example: 1
            400:
              description: Invalid filter format or query error
        """
        return super().get()

    @jwt_required()
    def post(self):
        """
        Create task type
        ---
        tags:
          - Crud
        description: Create a new task type with data provided in the
          request body. JSON format is expected. Task type names must
          be unique.
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - name
                  - for_entity
                properties:
                  name:
                    type: string
                    example: Animation
                  for_entity:
                    type: string
                    example: Shot
                  color:
                    type: string
                    example: "#FF5733"
        responses:
            201:
              description: Task type created successfully
              content:
                application/json:
                  schema:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        example: Animation
                      for_entity:
                        type: string
                        example: Shot
                      color:
                        type: string
                        example: "#FF5733"
                      created_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
            400:
              description: Invalid data format or task type name already exists
        """
        return super().post()

    def update_data(self, data):
        data = super().update_data(data)
        name = data.get("name", None)
        task_type = TaskType.get_by(name=name)
        if task_type is not None:
            raise WrongParameterException(
                "A task type with similar name already exists"
            )
        return data

    def post_creation(self, instance):
        tasks_service.clear_task_type_cache(str(instance.id))
        return instance.serialize()


class TaskTypeResource(BaseModelResource):
    def __init__(self):
        BaseModelResource.__init__(self, TaskType)

    def check_read_permissions(self, instance):
        return True

    @jwt_required()
    def get(self, instance_id):
        """
        Get task type
        ---
        tags:
          - Crud
        description: Retrieve a task type by its ID and return it as a
          JSON object. Supports including relations.
        parameters:
          - in: path
            name: instance_id
            required: true
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: relations
            required: false
            schema:
              type: boolean
            default: true
            example: true
            description: Whether to include relations
        responses:
            200:
              description: Task type retrieved successfully
              content:
                application/json:
                  schema:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        example: Animation
                      for_entity:
                        type: string
                        example: Shot
                      color:
                        type: string
                        example: "#FF5733"
                      created_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
            400:
              description: Invalid ID format or query error
        """
        return super().get(instance_id)

    @jwt_required()
    def put(self, instance_id):
        """
        Update task type
        ---
        tags:
          - Crud
        description: Update a task type with data provided in the
          request body. JSON format is expected. Task type names must
          be unique.
        parameters:
          - in: path
            name: instance_id
            required: true
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  name:
                    type: string
                    example: Updated Animation
                  color:
                    type: string
                    example: "#FF5734"
        responses:
            200:
              description: Task type updated successfully
              content:
                application/json:
                  schema:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        example: Updated Animation
                      for_entity:
                        type: string
                        example: Shot
                      color:
                        type: string
                        example: "#FF5734"
                      created_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T11:00:00Z"
            400:
              description: Invalid data format or task type name already exists
        """
        return super().put(instance_id)

    @jwt_required()
    def delete(self, instance_id):
        """
        Delete task type
        ---
        tags:
          - Crud
        description: Delete a task type by its ID. Returns empty
          response on success.
        parameters:
          - in: path
            name: instance_id
            required: true
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            204:
              description: Task type deleted successfully
            400:
              description: Integrity error or cannot delete
        """
        return super().delete(instance_id)

    def pre_delete(self, instance_dict):
        """
        Delete related ScheduleItems before deleting the TaskType.
        This runs AFTER check_delete_permissions, preventing foreign key
        constraint violations while maintaining security.
        """
        ScheduleItem.query.filter_by(task_type_id=instance_dict["id"]).delete()
        return instance_dict

    def update_data(self, data, instance_id):
        data = super().update_data(data, instance_id)
        name = data.get("name", None)
        if name is not None:
            task_type = TaskType.get_by(name=name)
            if task_type is not None and instance_id != str(task_type.id):
                raise WrongParameterException(
                    "A task type with similar name already exists"
                )
        return data

    def post_update(self, instance_dict, data):
        tasks_service.clear_task_type_cache(instance_dict["id"])
        return instance_dict

    def post_delete(self, instance_dict):
        tasks_service.clear_task_type_cache(instance_dict["id"])
        return instance_dict



from flask import request
from flask_restful import Resource, reqparse
from flask_jwt_extended import jwt_required

from zou.app.utils import permissions

from zou.app.services import (
    chats_service,
    entities_service,
    persons_service,
    user_service,
)
from zou.app.services.exception import WrongParameterException


class ChatResource(Resource):

    @jwt_required()
    def get(self, entity_id):
        """
        Get chat details
        ---
        description: Retrieve chat information and messages for a specific entity.
          Returns chat metadata including participants and all associated messages.
        tags:
          - Chat
        parameters:
          - in: path
            name: entity_id
            description: ID of the entity related to the chat
            type: string
            format: uuid
            required: true
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: Chat information successfully retrieved
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Chat unique identifier
                    entity_id:
                      type: string
                      format: uuid
                      description: Entity ID this chat is associated with
                    participants:
                      type: array
                      items:
                        type: string
                        format: uuid
                      description: List of participant user IDs
                    messages:
                      type: array
                      items:
                        type: object
                      description: Array of chat messages
        """
        entity = entities_service.get_entity(entity_id)
        user_service.check_project_access(entity["project_id"])
        user_service.check_entity_access(entity["id"])
        chat = chats_service.get_chat(entity_id)
        chat["messages"] = chats_service.get_chat_messages(entity_id)
        return chat


class ChatMessagesResource(Resource):

    @jwt_required()
    def get(self, entity_id):
        """
        Get chat messages
        ---
        description: Retrieve all chat messages for a specific entity.
          Returns a list of messages with sender information and timestamps.
        tags:
          - Chat
        parameters:
          - in: path
            name: entity_id
            description: ID of the entity related to the chat
            type: string
            format: uuid
            required: true
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: Chat messages successfully retrieved
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Message unique identifier
                      message:
                        type: string
                        description: Message content
                      person_id:
                        type: string
                        format: uuid
                        description: ID of the message sender
                      created_at:
                        type: string
                        format: date-time
                        description: Message creation timestamp
                      attachments:
                        type: array
                        items:
                          type: object
                        description: Array of file attachments
        """
        entity = entities_service.get_entity(entity_id)
        user_service.check_project_access(entity["project_id"])
        user_service.check_entity_access(entity["id"])
        return chats_service.get_chat_messages_for_entity(entity_id)

    @jwt_required()
    def post(self, entity_id):
        """
        Create chat message
        ---
        description: Create a new chat message for a specific entity.
          Supports both JSON and form data with optional file attachments.
          Only chat participants can send messages.
        tags:
          - Chat
        parameters:
          - in: path
            name: entity_id
            description: ID of the entity related to the chat
            type: string
            format: uuid
            required: true
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  message:
                    type: string
                    description: Message content to send
                    example: Hello, world!
                required:
                  - message
            multipart/form-data:
              schema:
                type: object
                properties:
                  message:
                    type: string
                    description: Message content to send
                    example: Hello, world!
                  files:
                    type: array
                    items:
                      type: string
                      format: binary
                    description: Files to attach to the message
                required:
                  - message
        responses:
          201:
            description: Chat message successfully created
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Created message unique identifier
                    message:
                      type: string
                      description: Message content
                    person_id:
                      type: string
                      format: uuid
                      description: ID of the message sender
                    created_at:
                      type: string
                      format: date-time
                      description: Message creation timestamp
                    attachments:
                      type: array
                      items:
                        type: object
                      description: Array of attached files
        """
        entity = entities_service.get_entity(entity_id)
        user_service.check_project_access(entity["project_id"])
        user_service.check_entity_access(entity["id"])

        person = persons_service.get_current_user()
        if request.is_json:
            location = ["values", "json"]
        else:
            location = ["values", "form"]
        parser = reqparse.RequestParser()
        parser.add_argument(
            "message", type=str, required=True, location=location
        )
        args = parser.parse_args()
        message = args["message"]
        files = request.files

        chat = chats_service.get_chat(entity_id)
        if person["id"] not in chat["participants"]:
            raise WrongParameterException(
                "You are not a participant of this chat"
            )

        return (
            chats_service.create_chat_message(
                chat["id"], person["id"], message, files=files
            ),
            201,
        )


class ChatMessageResource(Resource):

    @jwt_required()
    def get(self, entity_id, chat_message_id):
        """
        Get chat message
        ---
        description: Retrieve a specific chat message by its ID.
          Returns detailed message information including content and metadata.
        tags:
          - Chat
        parameters:
          - in: path
            name: entity_id
            description: ID of the entity related to the chat
            type: string
            format: uuid
            required: true
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: chat_message_id
            description: ID of the chat message
            type: string
            format: uuid
            required: true
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: Chat message successfully retrieved
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Message unique identifier
                    message:
                      type: string
                      description: Message content
                    person_id:
                      type: string
                      format: uuid
                      description: ID of the message sender
                    created_at:
                      type: string
                      format: date-time
                      description: Message creation timestamp
                    updated_at:
                      type: string
                      format: date-time
                      description: Message last update timestamp
                    attachments:
                      type: array
                      items:
                        type: object
                      description: Array of file attachments
        """
        entity = entities_service.get_entity(entity_id)
        user_service.check_project_access(entity["project_id"])
        user_service.check_entity_access(entity["id"])
        chat = chats_service.get_chat(entity_id)
        chat_message = chats_service.get_chat_message(chat_message_id)
        if chat_message["chat_id"] != chat["id"]:
            raise permissions.PermissionDenied()
        return chat_message

    @jwt_required()
    def delete(self, entity_id, chat_message_id):
        """
        Delete chat message
        ---
        description: Delete a specific chat message. Only the message author
          or administrators can delete messages.
        tags:
          - Chat
        parameters:
          - in: path
            name: entity_id
            description: ID of the entity related to the chat
            type: string
            format: uuid
            required: true
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: chat_message_id
            description: ID of the chat message to delete
            type: string
            format: uuid
            required: true
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          204:
            description: Chat message successfully deleted
        """
        entity = entities_service.get_entity(entity_id)
        user_service.check_project_access(entity["project_id"])
        user_service.check_entity_access(entity["id"])

        chat = chats_service.get_chat(entity_id)
        chat_message = chats_service.get_chat_message(chat_message_id)
        if chat_message["chat_id"] != chat["id"]:
            raise permissions.PermissionDenied()
        current_user = persons_service.get_current_user()
        if (
            chat_message["person_id"] != current_user["id"]
            and not permissions.has_admin_permissions()
        ):
            raise permissions.PermissionDenied()
        chats_service.delete_chat_message(chat_message_id)

        return "", 204



from sqlalchemy import create_engine, inspect, text
from sqlalchemy_utils import database_exists, create_database
from sqlalchemy.engine.url import URL
from sqlalchemy.orm import close_all_sessions


def get_db_uri():
    from zou.app.config import DATABASE

    return URL.create(**DATABASE).render_as_string(hide_password=False)


def reset_all():
    """
    Drop all tables then recreate them.
    """
    drop_all()
    return create_all()


def create_all():
    """
    Create all database tables.
    """
    from zou.app import db, config

    engine = create_engine(config.SQLALCHEMY_DATABASE_URI)
    if not database_exists(engine.url):
        create_database(engine.url)
    return db.create_all()


def drop_all():
    """
    Drop all database tables. Uses CASCADE to handle foreign keys from
    plugin tables that may not be tracked in db.metadata.
    """
    from zou.app import db

    db.session.flush()
    close_all_sessions()
    with db.engine.connect() as conn:
        conn.execute(text("DROP SCHEMA public CASCADE"))
        conn.execute(text("CREATE SCHEMA public"))
        conn.commit()


def is_init():
    """
    Check if database is initialized.
    """
    from zou.app import db
    from zou.app.models.project_status import ProjectStatus

    return (
        inspect(db.engine).has_table("person")
        and db.session.query(ProjectStatus).count() == 2
    )



from flask import abort, request
from flask_restful import Resource

from zou.app import config
from zou.app.models.person import Person
from zou.app.stores import config_store


class ConfigCheckResource(Resource):

    def get(self):
        token = request.headers.get("Authorization", "")
        if not token.startswith("Bearer ") or token[7:] != config.ADMIN_TOKEN:
            abort(403)

        comparison = config_store.get_config_comparison()
        comparison["active_users"] = Person.query.filter(
            Person.active, Person.is_bot.isnot(True)
        ).count()
        return comparison



from flask import current_app
from flask_jwt_extended import jwt_required

from sqlalchemy.exc import IntegrityError

from zou.app.models.project import Project
from zou.app.models.entity import Entity
from zou.app.models.entity_type import EntityType

from zou.app.blueprints.source.shotgun.base import (
    BaseImportShotgunResource,
    ImportRemoveShotgunBaseResource,
)

from zou.app.services import (
    assets_service,
    deletion_service,
    tasks_service,
    files_service,
    persons_service,
)

from zou.app.services.exception import AssetNotFoundException


class ImportShotgunAssetsResource(BaseImportShotgunResource):
    def __init__(self):
        BaseImportShotgunResource.__init__(self)

    @jwt_required()
    def post(self):
        """
        Import shotgun assets
        ---
        description: Import Shotgun assets. Send a list of Shotgun asset
          entries in the JSON body. Returns created or updated assets with
          parent-child relationships.
        tags:
          - Import
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: array
                items:
                  type: object
                  properties:
                    id:
                      type: integer
                      description: Shotgun ID of the asset
                      example: 12345
                    code:
                      type: string
                      description: Asset code
                      example: "Asset01"
                    description:
                      type: string
                      description: Asset description
                      example: "Main character asset"
                    sg_asset_type:
                      type: string
                      description: Asset type name
                      example: "Character"
                    project:
                      type: object
                      description: Project information
                      properties:
                        id:
                          type: integer
                          example: 11111
                    parents:
                      type: array
                      description: Parent assets
                      items:
                        type: object
                        properties:
                          id:
                            type: integer
                            example: 22222
              example:
                - id: 12345
                  code: "Asset01"
                  description: "Main character asset"
                  sg_asset_type: "Character"
                  project:
                    id: 11111
                  parents:
                    - id: 22222
        responses:
          200:
            description: Assets imported successfully
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Asset unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        description: Asset name
                        example: "Asset01"
                      description:
                        type: string
                        description: Asset description
                        example: "Main character asset"
                      created_at:
                        type: string
                        format: date-time
                        description: Creation timestamp
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        description: Update timestamp
                        example: "2024-01-15T11:00:00Z"
          400:
            description: Invalid request body or data format error
        """
        return super().post()

    def prepare_import(self):
        entity_type_names = self.extract_entity_type_names(self.sg_entries)
        assets_service.create_asset_types(entity_type_names)
        self.project_ids = Project.get_id_map()
        self.entity_type_ids = EntityType.get_id_map(field="name")
        self.parent_map = {}
        self.current_user_id = persons_service.get_current_user()["id"]

    def extract_entity_type_names(self, sg_assets):
        return {
            x["sg_asset_type"]
            for x in sg_assets
            if x["sg_asset_type"] is not None
        }

    def extract_data(self, sg_asset):
        entity_type_id = self.entity_type_ids.get(
            sg_asset["sg_asset_type"], None
        )
        project_id = self.project_ids.get(sg_asset["project"]["id"], None)
        parent_shotgun_ids = [
            asset["id"] for asset in sg_asset.get("parents", [])
        ]

        return {
            "name": sg_asset["code"],
            "shotgun_id": sg_asset["id"],
            "description": sg_asset["description"],
            "entity_type_id": entity_type_id,
            "project_id": project_id,
            "parent_shotgun_ids": parent_shotgun_ids,
        }

    def import_entry(self, data):
        entity = None
        parent_shotgun_ids = data["parent_shotgun_ids"]
        del data["parent_shotgun_ids"]

        try:
            entity = self.save_entity(data)
        except IntegrityError:
            current_app.logger.error(
                "Similar asset already exists "
                "or project is missing: %s" % data
            )

        if entity is not None:
            for parent_shotgun_id in parent_shotgun_ids:
                self.parent_map.setdefault(parent_shotgun_id, [])
                self.parent_map[parent_shotgun_id].append(
                    Entity.get(entity.id)
                )

        return entity

    def save_entity(self, data):
        entity = None
        try:
            entity = assets_service.get_raw_asset_by_shotgun_id(
                data["shotgun_id"]
            )
            entity.update(data)
            assets_service.clear_asset_cache(str(entity.id))
            current_app.logger.info("Entity updated: %s" % entity)
        except AssetNotFoundException:
            if data.get("entity_type_id", None) is not None:
                entity = Entity.create(**data, created_by=self.current_user_id)
                entity.save()
                current_app.logger.info("Entity created: %s" % entity)
            else:
                current_app.logger.info("Entity ignored: %s" % data["name"])
        return entity

    def post_processing(self):
        # We handle the fact that an asset can have multiple parents by using
        # the entities out field as a children field.
        for key in self.parent_map.keys():
            try:
                asset = assets_service.get_asset_by_shotgun_id(key)
                data = {"entities_out": self.parent_map[key]}
                assets_service.update_asset(asset["id"], data)
                assets_service.clear_asset_cache(asset["id"])
            except AssetNotFoundException:
                pass

        return self.parent_map


class ImportRemoveShotgunAssetResource(ImportRemoveShotgunBaseResource):
    def __init__(self):
        ImportRemoveShotgunBaseResource.__init__(
            self, Entity, self.delete_func
        )

    @jwt_required()
    def post(self):
        """
        Remove shotgun asset
        ---
        description: Remove a Shotgun asset from the database. Provide the
          Shotgun entry ID in the JSON body. If the asset has working files
          linked to tasks, it will be cancelled instead of deleted.
        tags:
          - Import
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - id
                properties:
                  id:
                    type: integer
                    description: Shotgun ID of the asset to remove
                    example: 12345
              example:
                id: 12345
        responses:
          200:
            description: Removal result returned
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    success:
                      type: boolean
                      description: Whether the removal was successful
                      example: true
                    removed_instance_id:
                      type: string
                      format: uuid
                      description: ID of the removed asset, if found
                      example: a24a6ea4-ce75-4665-a070-57453082c25
          400:
            description: Invalid request body or instance not found
        """
        return super().post()

    def delete_func(self, asset):
        try:
            asset = assets_service.get_asset_by_shotgun_id(asset.shotgun_id)
            tasks = tasks_service.get_tasks_for_asset(asset["id"])
            if self.is_working_files_linked(tasks):
                assets_service.cancel_asset(asset["id"])
            else:
                for task in tasks:
                    deletion_service.remove_task(task["id"])
                assets_service.remove_asset(asset["id"])
            return asset
        except AssetNotFoundException:
            return None

    def is_working_files_linked(self, tasks):
        is_working_files = False
        for task in tasks:
            working_files = files_service.get_working_files_for_task(
                task["id"]
            )
            if len(working_files) > 0:
                is_working_files = True
                break
        return is_working_files
